/*
 * Copyright (c) 2015 Cossack Labs Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "themis/sym_enc_message.h"

#include <string.h>

#include <soter/soter.h>

#include "themis/secure_cell_alg.h"

#define THEMIS_SYM_KDF_KEY_LABEL "Themis secure cell message key"
#define THEMIS_SYM_KDF_IV_LABEL "Themis secure cell message iv"

themis_status_t themis_sym_kdf(const uint8_t* master_key,
                               const size_t master_key_length,
                               const char* label,
                               const uint8_t* context,
                               const size_t context_length,
                               const uint8_t* context2,
                               const size_t context2_length,
                               uint8_t* key,
                               size_t key_length)
{
    THEMIS_CHECK_PARAM(master_key != NULL && master_key_length != 0);
    THEMIS_CHECK_PARAM(context != NULL && context_length != 0);
    soter_kdf_context_buf_t ctx[2] = {{context, context_length}, {context2, context2_length}};
    THEMIS_CHECK(soter_kdf(master_key,
                           master_key_length,
                           label,
                           ctx,
                           (context2 == NULL || context2_length == 0) ? 1 : 2,
                           key,
                           key_length)
                 == THEMIS_SUCCESS);
    return THEMIS_SUCCESS;
}

themis_status_t themis_auth_sym_plain_encrypt(uint32_t alg,
                                              const uint8_t* key,
                                              const size_t key_length,
                                              const uint8_t* iv,
                                              const size_t iv_length,
                                              const uint8_t* aad,
                                              const size_t aad_length,
                                              const uint8_t* message,
                                              const size_t message_length,
                                              uint8_t* encrypted_message,
                                              size_t* encrypted_message_length,
                                              uint8_t* auth_tag,
                                              size_t* auth_tag_length)
{
    soter_sym_ctx_t* ctx = soter_sym_aead_encrypt_create(alg, key, key_length, NULL, 0, iv, iv_length);
    THEMIS_CHECK(ctx != NULL);
    if (aad != NULL || aad_length != 0) {
        THEMIS_CHECK__(soter_sym_aead_encrypt_aad(ctx, aad, aad_length) == THEMIS_SUCCESS,
                       soter_sym_aead_encrypt_destroy(ctx);
                       return THEMIS_FAIL);
    }
    THEMIS_CHECK__(soter_sym_aead_encrypt_update(ctx, message, message_length, encrypted_message, encrypted_message_length)
                       == THEMIS_SUCCESS,
                   soter_sym_aead_encrypt_destroy(ctx);
                   return THEMIS_FAIL);
    THEMIS_CHECK__(soter_sym_aead_encrypt_final(ctx, auth_tag, auth_tag_length) == THEMIS_SUCCESS,
                   soter_sym_aead_encrypt_destroy(ctx);
                   return THEMIS_FAIL);
    soter_sym_aead_encrypt_destroy(ctx);
    return THEMIS_SUCCESS;
}

themis_status_t themis_auth_sym_plain_decrypt(uint32_t alg,
                                              const uint8_t* key,
                                              const size_t key_length,
                                              const uint8_t* iv,
                                              const size_t iv_length,
                                              const uint8_t* aad,
                                              const size_t aad_length,
                                              const uint8_t* encrypted_message,
                                              const size_t encrypted_message_length,
                                              uint8_t* message,
                                              size_t* message_length,
                                              const uint8_t* auth_tag,
                                              const size_t auth_tag_length)
{
    soter_sym_ctx_t* ctx = soter_sym_aead_decrypt_create(alg, key, key_length, NULL, 0, iv, iv_length);
    THEMIS_CHECK(ctx != NULL)
    if (aad != NULL || aad_length != 0) {
        THEMIS_CHECK__(soter_sym_aead_decrypt_aad(ctx, aad, aad_length) == THEMIS_SUCCESS,
                       soter_sym_aead_decrypt_destroy(ctx);
                       return THEMIS_FAIL);
    }
    THEMIS_CHECK__(soter_sym_aead_decrypt_update(ctx, encrypted_message, encrypted_message_length, message, message_length)
                       == THEMIS_SUCCESS,
                   soter_sym_aead_decrypt_destroy(ctx);
                   return THEMIS_FAIL);
    THEMIS_CHECK__(soter_sym_aead_decrypt_final(ctx, auth_tag, auth_tag_length) == THEMIS_SUCCESS,
                   soter_sym_aead_decrypt_destroy(ctx);
                   return THEMIS_FAIL);
    soter_sym_aead_decrypt_destroy(ctx);
    return THEMIS_SUCCESS;
}

themis_status_t themis_sym_plain_encrypt(uint32_t alg,
                                         const uint8_t* key,
                                         const size_t key_length,
                                         const uint8_t* iv,
                                         const size_t iv_length,
                                         const uint8_t* message,
                                         const size_t message_length,
                                         uint8_t* encrypted_message,
                                         size_t* encrypted_message_length)
{
    soter_sym_ctx_t* ctx = soter_sym_encrypt_create(alg, key, key_length, NULL, 0, iv, iv_length);
    THEMIS_CHECK__(ctx, return THEMIS_NO_MEMORY);
    size_t add_length = (*encrypted_message_length);
    themis_status_t res = soter_sym_encrypt_update(ctx,
                                                   message,
                                                   message_length,
                                                   encrypted_message,
                                                   encrypted_message_length);
    THEMIS_CHECK__(THEMIS_SUCCESS == res, soter_sym_encrypt_destroy(ctx); return res);
    add_length -= (*encrypted_message_length);
    res = soter_sym_encrypt_final(ctx, encrypted_message + (*encrypted_message_length), &add_length);
    THEMIS_CHECK__(res == THEMIS_SUCCESS, soter_sym_encrypt_destroy(ctx);
                   (*encrypted_message_length) += add_length;
                   return res);
    (*encrypted_message_length) += add_length;
    soter_sym_encrypt_destroy(ctx);
    return THEMIS_SUCCESS;
}

themis_status_t themis_sym_plain_decrypt(uint32_t alg,
                                         const uint8_t* key,
                                         const size_t key_length,
                                         const uint8_t* iv,
                                         const size_t iv_length,
                                         const uint8_t* encrypted_message,
                                         const size_t encrypted_message_length,
                                         uint8_t* message,
                                         size_t* message_length)
{
    soter_sym_ctx_t* ctx = soter_sym_decrypt_create(alg, key, key_length, NULL, 0, iv, iv_length);
    THEMIS_CHECK(ctx != NULL);
    size_t add_length = (*message_length);
    themis_status_t res = soter_sym_decrypt_update(ctx,
                                                   encrypted_message,
                                                   encrypted_message_length,
                                                   message,
                                                   message_length);
    THEMIS_CHECK__(THEMIS_SUCCESS == res, soter_sym_decrypt_destroy(ctx); return res);
    add_length -= (*message_length);
    res = soter_sym_decrypt_final(ctx, message + (*message_length), &add_length);
    THEMIS_CHECK__(res == THEMIS_SUCCESS, soter_sym_decrypt_destroy(ctx);
                   (*message_length) += add_length;
                   return res);
    (*message_length) += add_length;
    soter_sym_decrypt_destroy(ctx);
    return THEMIS_SUCCESS;
}

typedef struct themis_auth_sym_message_hdr_type {
    uint32_t alg;
    uint32_t iv_length;
    uint32_t auth_tag_length;
    uint32_t message_length;
} themis_auth_sym_message_hdr_t;

static const size_t auth_sym_context_length = sizeof(themis_auth_sym_message_hdr_t)
                                              + THEMIS_AUTH_SYM_IV_LENGTH
                                              + THEMIS_AUTH_SYM_AUTH_TAG_LENGTH;

themis_status_t themis_auth_sym_encrypt_message_(const uint8_t* key,
                                                 const size_t key_length,
                                                 uint8_t* derived_key,
                                                 const size_t derived_key_length,
                                                 const uint8_t* message,
                                                 const size_t message_length,
                                                 const uint8_t* in_context,
                                                 const size_t in_context_length,
                                                 uint8_t* out_context,
                                                 size_t* out_context_length,
                                                 uint8_t* encrypted_message,
                                                 size_t* encrypted_message_length)
{
    themis_auth_sym_message_hdr_t* hdr = (themis_auth_sym_message_hdr_t*)out_context;
    uint8_t* iv = out_context + sizeof(themis_auth_sym_message_hdr_t);
    uint8_t* auth_tag = iv + THEMIS_AUTH_SYM_IV_LENGTH;
    size_t auth_tag_length = THEMIS_AUTH_SYM_AUTH_TAG_LENGTH;

    *out_context_length = auth_sym_context_length;
    *encrypted_message_length = message_length;

    /*
     * We store the message length as 32-bit integer, messages longer than 4 GB are not supported.
     */
    if (message_length >= UINT32_MAX) {
        return THEMIS_INVALID_PARAMETER;
    }
    THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                       key_length,
                                       THEMIS_SYM_KDF_KEY_LABEL,
                                       (uint8_t*)(&message_length),
                                       sizeof(uint32_t),
                                       in_context,
                                       in_context_length,
                                       derived_key,
                                       derived_key_length),
                        THEMIS_SUCCESS);

    THEMIS_CHECK(soter_rand(iv, THEMIS_AUTH_SYM_IV_LENGTH) == SOTER_SUCCESS);

    hdr->alg = THEMIS_AUTH_SYM_ALG;
    hdr->iv_length = THEMIS_AUTH_SYM_IV_LENGTH;
    hdr->auth_tag_length = THEMIS_AUTH_SYM_AUTH_TAG_LENGTH;
    hdr->message_length = (uint32_t)message_length;

    THEMIS_STATUS_CHECK(themis_auth_sym_plain_encrypt(THEMIS_AUTH_SYM_ALG,
                                                      derived_key,
                                                      derived_key_length,
                                                      iv,
                                                      THEMIS_AUTH_SYM_IV_LENGTH,
                                                      in_context,
                                                      in_context_length,
                                                      message,
                                                      message_length,
                                                      encrypted_message,
                                                      encrypted_message_length,
                                                      auth_tag,
                                                      &auth_tag_length),
                        THEMIS_SUCCESS);

    THEMIS_CHECK(auth_tag_length == THEMIS_AUTH_SYM_AUTH_TAG_LENGTH);

    return THEMIS_SUCCESS;
}

themis_status_t themis_auth_sym_encrypt_message(const uint8_t* key,
                                                const size_t key_length,
                                                const uint8_t* message,
                                                const size_t message_length,
                                                const uint8_t* in_context,
                                                const size_t in_context_length,
                                                uint8_t* out_context,
                                                size_t* out_context_length,
                                                uint8_t* encrypted_message,
                                                size_t* encrypted_message_length)
{
    themis_status_t status = THEMIS_FAIL;
    uint8_t derived_key[THEMIS_AUTH_SYM_KEY_LENGTH / 8];

    THEMIS_CHECK_PARAM(key != NULL && key_length != 0);
    THEMIS_CHECK_PARAM(message != NULL && message_length != 0);
    /* in_context may be NULL and empty */
    THEMIS_CHECK_PARAM(out_context_length != NULL);
    THEMIS_CHECK_PARAM(encrypted_message_length != NULL);

    if (!out_context || !encrypted_message || (*out_context_length < auth_sym_context_length)
        || (*encrypted_message_length < message_length)) {
        *out_context_length = auth_sym_context_length;
        *encrypted_message_length = message_length;
        return THEMIS_BUFFER_TOO_SMALL;
    }

    status = themis_auth_sym_encrypt_message_(key,
                                              key_length,
                                              derived_key,
                                              sizeof(derived_key),
                                              message,
                                              message_length,
                                              in_context,
                                              in_context_length,
                                              out_context,
                                              out_context_length,
                                              encrypted_message,
                                              encrypted_message_length);

    soter_wipe(derived_key, sizeof(derived_key));

    return status;
}

themis_status_t themis_auth_sym_decrypt_message_(const uint8_t* key,
                                                 const size_t key_length,
                                                 uint8_t* derived_key,
                                                 const size_t derived_key_length,
                                                 const uint8_t* in_context,
                                                 const size_t in_context_length,
                                                 const uint8_t* context,
                                                 const size_t context_length,
                                                 const uint8_t* encrypted_message,
                                                 const size_t encrypted_message_length,
                                                 uint8_t* message,
                                                 size_t* message_length)
{
    themis_auth_sym_message_hdr_t* hdr = (themis_auth_sym_message_hdr_t*)context;
    const uint8_t* iv = context + sizeof(themis_auth_sym_message_hdr_t);
    const uint8_t* auth_tag = iv + hdr->iv_length;
    UNUSED(context_length);

    *message_length = hdr->message_length;

    /*
     * We store the message length as 32-bit integer, messages longer than 4 GB are not supported.
     */
    if (encrypted_message_length >= UINT32_MAX) {
        return THEMIS_INVALID_PARAMETER;
    }
    THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                       key_length,
                                       THEMIS_SYM_KDF_KEY_LABEL,
                                       (uint8_t*)(&encrypted_message_length),
                                       sizeof(uint32_t),
                                       in_context,
                                       in_context_length,
                                       derived_key,
                                       derived_key_length),
                        THEMIS_SUCCESS);

    THEMIS_STATUS_CHECK(themis_auth_sym_plain_decrypt(hdr->alg,
                                                      derived_key,
                                                      derived_key_length,
                                                      iv,
                                                      hdr->iv_length,
                                                      in_context,
                                                      in_context_length,
                                                      encrypted_message,
                                                      hdr->message_length,
                                                      message,
                                                      message_length,
                                                      auth_tag,
                                                      hdr->auth_tag_length),
                        THEMIS_SUCCESS);

    return THEMIS_SUCCESS;
}

#ifdef SCELL_COMPAT
themis_status_t themis_auth_sym_decrypt_message_compat(const uint8_t* key,
                                                       const size_t key_length,
                                                       uint8_t* derived_key,
                                                       const size_t derived_key_length,
                                                       const uint8_t* in_context,
                                                       const size_t in_context_length,
                                                       const uint8_t* context,
                                                       const size_t context_length,
                                                       const uint8_t* encrypted_message,
                                                       const size_t encrypted_message_length,
                                                       uint8_t* message,
                                                       size_t* message_length)
{
    themis_auth_sym_message_hdr_t* hdr = (themis_auth_sym_message_hdr_t*)context;
    const uint8_t* iv = context + sizeof(themis_auth_sym_message_hdr_t);
    const uint8_t* auth_tag = iv + hdr->iv_length;
    UNUSED(context_length);

    *message_length = hdr->message_length;

    /*
     * Note that we use sizeof(uint64_t) here as that's the size used by buggy Themis 0.9.6.
     */
    THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                       key_length,
                                       THEMIS_SYM_KDF_KEY_LABEL,
                                       (uint8_t*)(&encrypted_message_length),
                                       sizeof(uint64_t),
                                       in_context,
                                       in_context_length,
                                       derived_key,
                                       derived_key_length),
                        THEMIS_SUCCESS);

    THEMIS_STATUS_CHECK(themis_auth_sym_plain_decrypt(hdr->alg,
                                                      derived_key,
                                                      derived_key_length,
                                                      iv,
                                                      hdr->iv_length,
                                                      in_context,
                                                      in_context_length,
                                                      encrypted_message,
                                                      hdr->message_length,
                                                      message,
                                                      message_length,
                                                      auth_tag,
                                                      hdr->auth_tag_length),
                        THEMIS_SUCCESS);

    return THEMIS_SUCCESS;
}
#endif

themis_status_t themis_auth_sym_decrypt_message(const uint8_t* key,
                                                const size_t key_length,
                                                const uint8_t* in_context,
                                                const size_t in_context_length,
                                                const uint8_t* context,
                                                const size_t context_length,
                                                const uint8_t* encrypted_message,
                                                const size_t encrypted_message_length,
                                                uint8_t* message,
                                                size_t* message_length)
{
    themis_status_t status = THEMIS_FAIL;
    themis_auth_sym_message_hdr_t* hdr = (themis_auth_sym_message_hdr_t*)context;
    uint8_t derived_key[THEMIS_AUTH_SYM_KEY_LENGTH / 8];

    THEMIS_CHECK_PARAM(key != NULL && key_length != 0);
    /* in_context may be NULL and empty */
    THEMIS_CHECK_PARAM(context != NULL && context_length >= sizeof(themis_auth_sym_message_hdr_t));
    THEMIS_CHECK_PARAM(message_length != NULL);

    if (!message || (*message_length < hdr->message_length)) {
        *message_length = hdr->message_length;
        return THEMIS_BUFFER_TOO_SMALL;
    }

    /*
     * encrypted_message and encrypted_message_length may be NULL in some cases
     * when plaintext length is queried. Check encrypted message after returning
     * THEMIS_BUFFER_TOO_SMALL. context length is in a similar situation.
     */
    THEMIS_CHECK_PARAM(encrypted_message != NULL);
    THEMIS_CHECK_PARAM(encrypted_message_length >= hdr->message_length);
    THEMIS_CHECK_PARAM(context_length >= (sizeof(themis_auth_sym_message_hdr_t) + hdr->iv_length
                                          + hdr->auth_tag_length));

    status = themis_auth_sym_decrypt_message_(key,
                                              key_length,
                                              derived_key,
                                              sizeof(derived_key),
                                              in_context,
                                              in_context_length,
                                              context,
                                              context_length,
                                              encrypted_message,
                                              encrypted_message_length,
                                              message,
                                              message_length);
    /*
     * Workaround for compatibility with Themis 0.9.6 on 64-bit systems.
     */
#ifdef SCELL_COMPAT
    if (status != THEMIS_SUCCESS && sizeof(size_t) == sizeof(uint64_t)) {
        status = themis_auth_sym_decrypt_message_compat(key,
                                                        key_length,
                                                        derived_key,
                                                        sizeof(derived_key),
                                                        in_context,
                                                        in_context_length,
                                                        context,
                                                        context_length,
                                                        encrypted_message,
                                                        encrypted_message_length,
                                                        message,
                                                        message_length);
    }
#endif

    soter_wipe(derived_key, sizeof(derived_key));

    return status;
}

typedef struct themis_sym_message_hdr_type {
    uint32_t alg;
    uint32_t iv_length;
    uint32_t message_length;
} themis_sym_message_hdr_t;

themis_status_t themis_sym_encrypt_message_u_(const uint8_t* key,
                                              const size_t key_length,
                                              const uint8_t* message,
                                              const size_t message_length,
                                              const uint8_t* context,
                                              const size_t context_length,
                                              uint8_t* encrypted_message,
                                              size_t* encrypted_message_length)
{
    THEMIS_CHECK_PARAM(message != NULL && message_length != 0);
    THEMIS_CHECK_PARAM(context != NULL && context_length != 0);
    if ((*encrypted_message_length) < message_length) {
        (*encrypted_message_length) = message_length;
        return THEMIS_BUFFER_TOO_SMALL;
    }
    (*encrypted_message_length) = message_length;
    uint8_t iv[THEMIS_SYM_IV_LENGTH];
    THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                       key_length,
                                       THEMIS_SYM_KDF_IV_LABEL,
                                       context,
                                       context_length,
                                       NULL,
                                       0,
                                       iv,
                                       THEMIS_SYM_IV_LENGTH),
                        THEMIS_SUCCESS);
    THEMIS_STATUS_CHECK(themis_sym_plain_encrypt(THEMIS_SYM_ALG,
                                                 key,
                                                 key_length,
                                                 iv,
                                                 THEMIS_SYM_IV_LENGTH,
                                                 message,
                                                 message_length,
                                                 encrypted_message,
                                                 encrypted_message_length),
                        THEMIS_SUCCESS);
    return THEMIS_SUCCESS;
}

themis_status_t themis_sym_encrypt_message_u(const uint8_t* key,
                                             const size_t key_length,
                                             const uint8_t* context,
                                             const size_t context_length,
                                             const uint8_t* message,
                                             const size_t message_length,
                                             uint8_t* encrypted_message,
                                             size_t* encrypted_message_length)
{
    uint8_t key_[THEMIS_SYM_KEY_LENGTH / 8];

    // TODO: TYPE WARNING Should update `sizeof(uint32_t)` to `sizeof(message_length)` after
    // changing encrypted_message_length type to uint32_t
    THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                       key_length,
                                       THEMIS_SYM_KDF_KEY_LABEL,
                                       (uint8_t*)(&message_length),
                                       sizeof(uint32_t),
                                       NULL,
                                       0,
                                       key_,
                                       sizeof(key_)),
                        THEMIS_SUCCESS);
    return themis_sym_encrypt_message_u_(key_,
                                         sizeof(key_),
                                         message,
                                         message_length,
                                         context,
                                         context_length,
                                         encrypted_message,
                                         encrypted_message_length);
}

themis_status_t themis_sym_decrypt_message_u_(const uint8_t* key,
                                              const size_t key_length,
                                              const uint8_t* context,
                                              const size_t context_length,
                                              const uint8_t* encrypted_message,
                                              const size_t encrypted_message_length,
                                              uint8_t* message,
                                              size_t* message_length)
{
    THEMIS_CHECK_PARAM(encrypted_message != NULL && encrypted_message_length != 0);
    THEMIS_CHECK_PARAM(context != NULL && context_length != 0);
    if ((*message_length) < encrypted_message_length) {
        (*message_length) = encrypted_message_length;
        return THEMIS_BUFFER_TOO_SMALL;
    }
    (*message_length) = encrypted_message_length;
    uint8_t iv[THEMIS_SYM_IV_LENGTH];
    THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                       key_length,
                                       THEMIS_SYM_KDF_IV_LABEL,
                                       context,
                                       context_length,
                                       NULL,
                                       0,
                                       iv,
                                       THEMIS_SYM_IV_LENGTH),
                        THEMIS_SUCCESS);
    THEMIS_STATUS_CHECK(themis_sym_plain_decrypt(THEMIS_SYM_ALG,
                                                 key,
                                                 key_length,
                                                 iv,
                                                 THEMIS_SYM_IV_LENGTH,
                                                 encrypted_message,
                                                 encrypted_message_length,
                                                 message,
                                                 message_length),
                        THEMIS_SUCCESS);
    return THEMIS_SUCCESS;
}

themis_status_t themis_sym_decrypt_message_u(const uint8_t* key,
                                             const size_t key_length,
                                             const uint8_t* context,
                                             const size_t context_length,
                                             const uint8_t* encrypted_message,
                                             const size_t encrypted_message_length,
                                             uint8_t* message,
                                             size_t* message_length)
{
    uint8_t key_[THEMIS_SYM_KEY_LENGTH / 8];

    // TODO: TYPE WARNING Should update `sizeof(uint32_t)` to `sizeof(encrypted_message_length)`
    // after changing encrypted_message_length type to uint32_t
    THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                       key_length,
                                       THEMIS_SYM_KDF_KEY_LABEL,
                                       (uint8_t*)(&encrypted_message_length),
                                       sizeof(uint32_t),
                                       NULL,
                                       0,
                                       key_,
                                       sizeof(key_)),
                        THEMIS_SUCCESS);
    themis_status_t decryption_result = themis_sym_decrypt_message_u_(key_,
                                                                      sizeof(key_),
                                                                      context,
                                                                      context_length,
                                                                      encrypted_message,
                                                                      encrypted_message_length,
                                                                      message,
                                                                      message_length);

#ifdef SCELL_COMPAT

    // we are on x64, should sizeof(uin64_t) for backwards compatibility with themis 0.9.6 x64
    if (sizeof(size_t) == sizeof(uint64_t)) {
        if (decryption_result != THEMIS_SUCCESS && decryption_result != THEMIS_BUFFER_TOO_SMALL) {
            // TODO: TYPE WARNING `sizeof(uint64_t)`. Fix that on next versions
            THEMIS_STATUS_CHECK(themis_sym_kdf(key,
                                               key_length,
                                               THEMIS_SYM_KDF_KEY_LABEL,
                                               (uint8_t*)(&encrypted_message_length),
                                               sizeof(uint64_t),
                                               NULL,
                                               0,
                                               key_,
                                               sizeof(key_)),
                                THEMIS_SUCCESS);
            decryption_result = themis_sym_decrypt_message_u_(key_,
                                                              sizeof(key_),
                                                              context,
                                                              context_length,
                                                              encrypted_message,
                                                              encrypted_message_length,
                                                              message,
                                                              message_length);
        }
    }
#endif

    return decryption_result;
}
